///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//        Program Name:  common.h                                            //
//              Author:  Mike McCollister ( 2000-2009)                      //
//            Language:  PRC-Tools 2.3                                       //
//                       Palm OS 5 SDK R3                                    //
//                       PilRC 3.2                                           //
//                       HandEra Header Files from SDK 1.04                  //
//                       palmOne Header Files from SDK version 5.1           //
//                       Sony Header Files from SDK 5.0 version 1.2          //
//               Notes:  Distributed under the GNU GPL.                      //
//                       Based on LinkHistoryHack by Ben Darnell             //
//                       <bgdarnel@unity.ncsu.edu> which was based on        //
//                       SampleHack by Roger Chaplin                         //
//                       <foursquaredev@att.net>.                            //
//         Description:  Common code header file for McPhling.               //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

#ifndef COMMON_H
#define	COMMON_H 1

///////////////////
// Include Files //
///////////////////

#include <PalmOS.h>
#include <Extensions/ExpansionMgr/VFSMgr.h>

#include <HandEra/Vga.h>

#include "McPhling.h"

/////////////
// Defines //
/////////////

#define BEEP			SndPlaySystemSound(sndInfo)
#define CLICK			SndPlaySystemSound(sndClick)
#define DEBUG_BEEP		BEEP

#ifndef SetBits
#define SetBits(b, len)         (((1U << ((len) - 1)) - 1U +  \
                                  (1U << ((len) - 1))) << (b))
#endif

// pinMaxConstraintSize is set to the max value of a Coord such that it can
// grow as big as the hardware supports
#ifndef pinMaxConstraintSize
#define pinMaxConstraintSize 	SetBits(0, (sizeof(Coord) * 8) - 1)
#endif

/////////////////////////
// Function Prototypes //
/////////////////////////

void GetActivation(McActivation *pMpActivation);
Boolean GetIconName(UInt16 cardNo, LocalID dbID, Char *pIconChar);
void GetMRUApps(McPhlingPrefs *mpPrefs, JumpRefPref *jumpRefs);
void GetPrefsAndNewIgnored(McPhlingPrefs *pMpPrefs);
void *Id2Ptr(UInt16 wID);
void LstSetBestWidth(UInt16 listID);
void DisplayWaitForm(Boolean bVgaExists, VgaScreenModeType *pScreenMode,
                     VgaRotateModeType *pRotateMode, FormPtr *pNewForm);
void RemoveWaitForm(Boolean bVgaExists, VgaScreenModeType screenMode,
                    VgaRotateModeType rotateMode, FormPtr pNewForm,
                    FormPtr pForm);
UInt16 GetROMVer(UInt16 *ROMVerMajor, UInt16 *ROMVerMinor);
void VFSSetup(VFSInfo *vfsInfo);
void EnableOS5(Boolean bEnabled);
Boolean GetCollapseGraffiti(Boolean *pbPS, Boolean *pbSony);
void CollapseSetUserState(FormPtr pForm, Boolean bForce);

#endif
